import Base from './base';
import { getActualPoint, getClickPoint, createSvg, getHtmlToText, keepLastIndex, getInitialPoint } from './utils';
import { onClickOutside } from '../ui_utils'

export default class Text extends Base {
  constructor ({
    container,
    annotation,
    page,
    viewport,
    scale,
    eventBus,
    layer,
    show = false
  }) {
    super({
      container,
      annotation,
      page,
      viewport,
      scale,
      eventBus,
      show
    })

    this.textNoteStr = `<svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none">
      <path fill-rule="evenodd" clip-rule="evenodd" d="M2.39961 3.60004H21.5996V18H11.9996L5.99961 21.6V18H2.39961V3.60004Z" fill="currentColor"/>
      <rect x="5.99902" y="7.20001" width="7.2" height="1.8" fill="white"/>
      <rect x="5.99902" y="11.4" width="12" height="1.8" fill="white"/>
    </svg>`
    this.deleteSvgStr = `<path xmlns="http://www.w3.org/2000/svg" fill-rule="evenodd" clip-rule="evenodd" d="M14 1.5V4.5H19V5.5H16.5V18.5H3.5V5.5H1V4.5H6V1.5H14ZM4.5 5.5V17.5H15.5V5.5H4.5ZM13 2.5V4.5H7V2.5H13ZM8.5 8V15H7.5V8H8.5ZM12.5 15V8H11.5V15H12.5Z" fill="#333333"/>`
    
    this.layer = layer
    this.hidden = true
    this.initial = false
    this.outline = null

    this.start = null
    this.end = null

    this.newStart = null
    this.newEnd = null

    this.startCircle = null
    this.endCircle = null

    this.show = show

    this.popShow = false

    this.onClick = this.handleClick.bind(this)
    this.onBlur = this.handleBlur.bind(this)
    this.onDelete = this.handleDelete.bind(this)
    
    this.onMousedown = this.handleMouseDown.bind(this)
    this.onMousemove = this.handleMouseMove.bind(this)
    this.onMouseup = this.handleMouseUp.bind(this)
    this.render()
  }

  render () {
    this.renderTextElement()
    this.renderTextEditor()
    
    this.container.append(this.annotationContainer)
  }

  handleDelete (event) {
    if (this.layer.tool) {
      event.stopPropagation()
    }
    this.annotationContainer.remove()
    this.annotation.isDelete = true
    this.eventBus.dispatch('annotationChange', {
      type: 'delete',
      annotation: {
        operate: "del-annot",
        name: this.annotation.name,
        page: this.page
      }
    })
  }

  handleBlur () {
    const annotation = this.annotation
    const contents = getHtmlToText(this.textEditorElement)
    if (contents !== annotation.contents) {
      this.eventBus.dispatch('annotationChange', {
        type: 'modify',
        annotation: {
          operate: "mod-annot",
          name: annotation.name,
          page: this.page,
          content: contents,
          contents
        }
      })
      annotation.contents = annotation.content = contents
    }

    this.textEditorContainer.remove()
  }

  handleMouseDown (event) {
    if (this.layer.addMode) return
    if (this.layer.tool) {
      event.stopPropagation()
    }
    this.moving = false
    const { pageX, pageY } = getClickPoint(event)
    this.startState = {
      clickX: pageX,
      clickY: pageY,
    }
    document.addEventListener('mousemove', this.onMousemove)
    document.addEventListener('mouseup', this.onMouseup)
    document.addEventListener('touchmove', this.onMousemove)
    document.addEventListener('touchend', this.onMouseup)
  };

  handleMouseMove (event) {
    if (event.type === 'touchmove') {
      document.body.style.overscrollBehavior = 'none';
    }
    const { pageX, pageY } = getClickPoint(event)
    if (pageX === this.startState.clickX && pageY === this.startState.clickY) return
    this.moving = true
    const start = this.start
    const end = this.end
    const startState = this.startState

    const { width, height } = this.viewport

    let x1 = pageX - (startState.clickX - this.start.x);
    let y1 = pageY - (startState.clickY - this.start.y);
    let x2 = pageX - (startState.clickX - this.end.x);
    let y2 = pageY - (startState.clickY - this.end.y);

    const rect = {
      width: Math.abs(start.x -end.x),
      height: Math.abs(start.y -end.y)
    }
    if (x1 < x2) {
      x1 = Math.max(0, x1)
      x1 = Math.min(width - rect.width, x1)
      x2 = Math.max(rect.width, x2)
      x2 = Math.min(width, x2)
    } else {
      x2 = Math.max(0, x2)
      x2 = Math.min(width - rect.width, x2)
      x1 = Math.max(rect.width, x1)
      x1 = Math.min(width, x1)
    }

    if (y1 < y2) {
      y1 = Math.max(0, y1)
      y1 = Math.min(height - rect.height, y1)
      y2 = Math.max(rect.height, y2)
      y2 = Math.min(height, y2)
    } else {
      y2 = Math.max(0, y2)
      y2 = Math.min(height - rect.height, y2)
      y1 = Math.max(rect.height, y1)
      y1 = Math.min(height, y1)
    }

    this.update({
      start: { x: x1, y: y1 },
      end: { x: x2, y: y2 },
    });
  }

  handleMouseUp (event) {
    if (this.layer.tool) {
      event.stopPropagation()
    }
    if (event.type === 'touchend') {
      document.body.style.overscrollBehavior = 'auto';
    }
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)

    const { pageX, pageY } = getClickPoint(event)
    if (pageX === this.startState.clickX && pageY === this.startState.clickY) return

    this.start = this.newStart
    this.end = this.newEnd

    const annotation = this.annotation
    const start = getInitialPoint(this.start, this.viewport, this.viewport.scale)
    const end = getInitialPoint(this.end, this.viewport, this.viewport.scale)

    const linePoints = start.x + ',' + start.y + ',' + end.x + ',' + end.y
    if (linePoints === annotation.linePoints) return
    annotation.linePoints = linePoints

    this.eventBus.dispatch('annotationChange', {
      type: 'modify',
      annotation: {
        operate: "mod-annot",
        name: annotation.name,
        page: this.page,
        rect: linePoints
      }
    })
  }

  update ({ start, end }) {
    const rect = this.rectCalc({start, end})

    this.newStart = start
    this.newEnd = end

    this.annotationContainer.style.top = rect.top + 'px'
    this.annotationContainer.style.left = rect.left + 'px'
  }

  rectCalc ({start, end}) {
    return {
      top: start.y < end.y ? start.y : end.y,
      left: start.x < end.x ? start.x : end.x,
      width: Math.abs(end.x - start.x),
      height: Math.abs(end.y - start.y),
    };
  };

  getActualRect (viewport, s,) {
    const annotation = this.annotation
    const [x1, y1, x2, y2] = annotation.rect.split(',')
    
    const start = getActualPoint(
      {
        x: x1,
        y: y1
      },
      viewport,
      s
    )
    const end = getActualPoint(
      {
        x: x2,
        y: y2
      },
      viewport,
      s
    )
    return {
      start,
      end
    }
  }

  handleClickOutside () {
    this.popShow = false
    this.textEditorContainer.remove()
    this.handleBlur()
  }

  setCss (ele, cssText) {
    if (!ele) return
    if (cssText) {
      for (let key in cssText) {
        ele.style[key] = cssText[key]
      }
    }
  }

  handleClick (event) {
    if (this.layer.addMode) return
    if (this.layer.tool) {
      event.stopPropagation()
    }
    if (this.moving) return
    if (this.popShow) return
    this.popShow = true
    this.annotationContainer.append(this.textEditorContainer)
    this.textEditorElement.focus()
    keepLastIndex(this.textEditorElement)
    onClickOutside([this.annotationContainer], this.handleClickOutside.bind(this))
  }

  updateTool () {
    if (!this.popShow) {
      if (this.layer.selectedElementName === this.annotation.name) {
        this.layer.selectedElementName = null
      }
      this.outerLineContainer.remove()
    } else {
      if (this.layer.selectedElementName !== this.annotation.name) {
        this.layer.selectedElementName = this.annotation.name
      }
      this.container.append(this.outerLineContainer)
      this.outerLine.addEventListener('mousedown', this.onMousedown)
      this.outerLine.addEventListener('touchstart', this.onMousedown)
    }
  }

  renderTextElement () {
    const { start, end } = this.getActualRect(
      this.viewport,
      this.viewport.scale
    )
    this.start = start
    this.end = end

    const annotation = this.annotation

    const rect = this.rectCalc({start, end})

    let annotationContainer = document.createElement('div')
    annotationContainer.id = annotation.name
    annotationContainer.className = 'annotation text'
    this.setCss(annotationContainer, {
      position: 'absolute',
      left: rect.left + 'px',
      top: rect.top + 'px',
      width: '24px',
      height: '24px',
      zIndex: 6,
      color: '#FF0000'
    })
    annotationContainer.innerHTML = this.textNoteStr
    annotationContainer.addEventListener('click', this.onClick)
    annotationContainer.addEventListener('mousedown', this.onMousedown)
    this.annotationContainer = annotationContainer
  }

  renderTextEditor () {
    const textEditorContainer = document.createElement('div')
    this.textEditorContainer = textEditorContainer
    this.setCss(textEditorContainer, {
      position: 'absolute',
      left: '16px',
      top: '28px',
      border: '2px solid #FF0000',
      outline: 'none',
      padding: '0 8px 28px',
      borderRadius: '4px',
      background: '#FFFFFF'
    })

    const title = document.createElement('div')
    title.className = 'text-title'
    title.innerText = 'Note'

    textEditorContainer.append(title)

    let textEditorElement = document.createElement('div')
    textEditorElement.setAttribute('role', 'textbox')
    textEditorElement.setAttribute('contenteditable', true)
    textEditorElement.setAttribute('spellcheck', false)
    this.setCss(textEditorElement, {
      width: '290px',
      height: '110px',
      fontSize: '16px',
      color: '#000',
      lineHeight: '1.2em',
      outline: 'none',
      overflow: 'auto',
      whiteSpace: 'pre-wrap',
      textAlign: 'left'
    })
    textEditorElement.className = 'text-editor'

    if (this.annotation.content) {
      textEditorElement.innerHTML = this.annotation.content
    } else if (this.annotation.contents) {
      textEditorElement.innerHTML = this.annotation.contents
    } else if (this.annotation['contents-richtext']) {
      const parser = new window.DOMParser();
      const xmlDoc = parser.parseFromString(this.annotation['contents-richtext'], 'text/xml');
      const error = xmlDoc.getElementsByTagName("parsererror")
      if (error.length > 0) {
        textEditorElement.innerHTML = this.annotation['contents-richtext']
      } else {
        textEditorElement.innerHTML = xmlDoc.firstElementChild.innerText
      }
    }
    
    this.textEditorElement = textEditorElement

    textEditorContainer.append(textEditorElement)

    const deletetButton = createSvg(
      "svg",
      {
        width: "20",
        height: "20",
        viewBox: "0 0 20 20",
        fill: "none",
        class: 'delete-button'
      },
      {
        width: '20px',
        height: '20px',
        position: 'absolute',
        right: '10px',
        bottom: '10px'
      }
    );
    deletetButton.innerHTML = this.deleteSvgStr
    this.deletetButton = deletetButton
    textEditorContainer.append(deletetButton)
    this.deletetButton.addEventListener('click', this.onDelete)
  }
}
