class MarkupAnnotation {
  constructor ({
    eventBus,
    viewport,
    pages,
    scale
  }) {
    this.eventBus = eventBus
    this.viewport = viewport
    this.tool = ''
    this.pages = pages
    this.scale = scale
    this.onHandleUp = this.handleUp.bind(this)
    this.onHandleTool = this.handleTool.bind(this)

    this.init()
  }

  init () {
    document.addEventListener('selectionchange', this.handleSelectChange)
    this.eventBus._on('toolChanged', this.onHandleTool)
  }

  handleSelectChange () {
    const selection = document.getSelection()
    const container = document.querySelector('.document')
    if (!selection?.isCollapsed) {
      container.classList.add('selection')
    } else {
      container.classList.remove('selection')
    }
  }

  handleTool ({
    tool,
    color
  }) {
    this.tool = tool
    this.color = color

    const markupType = ['highlight', 'underline', 'squiggly', 'strikeout', '']
    if (markupType.includes(this.tool)) {
      document.body.style.userSelect = 'auto'
      document.body.style.webkitUserSelect = 'auto'
    } else {
      document.body.style.userSelect = 'none'
      document.body.style.webkitUserSelect = 'none'
    }
  }

  get tool () {
    return this._tool
  }

  set tool (toolType) {
    if (toolType === this._tool) return
    this._tool = toolType
    if (!toolType) {
      document.removeEventListener('mouseup', this.onHandleUp)
      document.removeEventListener('touchend', this.onHandleUp)
      return
    }
    document.addEventListener('mouseup', this.onHandleUp)
    document.addEventListener('touchend', this.onHandleUp)
  }

  destroy () {
    this.eventBus._off('toolChanged', this.onHandleTool)
    document.removeEventListener('mouseup', this.onHandleUp)
    document.removeEventListener('touchend', this.onHandleUp)
  }

  handleUp (event) {
    const selection = document.getSelection()
    let textLayer = ''
    if (!selection?.isCollapsed) {
      if (event.target) {
        textLayer = (event.target).parentNode

        if (!textLayer.classList.contains('textLayer')) {
          textLayer = textLayer.closest('.page').querySelector('.textLayer')
        }
        const markupType = ['highlight', 'underline', 'squiggly', 'strikeout']
        if (textLayer && (markupType.includes(this.tool))) {
          const newMarkup = this.getMarkupWithSelection({
            textLayer,
            content: selection && selection.toString(),
          })

          if (newMarkup) {
            this.eventBus.dispatch('annotationChange', {
              type: 'add',
              annotation: newMarkup
            })
            const pageView = this.pages[newMarkup.page]
            if (pageView && pageView.compdfAnnotationLayer) {
              pageView.compdfAnnotationLayer.renderAnnotation(newMarkup)
            }
            selection.removeAllRanges()
          }
        }
      }
    }
  }

  getMarkupWithSelection ({
    textLayer,
    content,
  }) {
    const selection = document.getSelection()
  
    if (!selection || selection.isCollapsed) return null
    const {
      startContainer,
      startOffset,
      endContainer,
      endOffset,
    } = selection.getRangeAt(0)
    const startElement = startContainer.parentNode
    const endElement = endContainer.parentNode
    const startPage = startElement?.closest('.page')
    const startPageNum = parseInt(
      startPage.getAttribute('data-page-number'),
      10,
    )
  
    const pageHeight = startPage.offsetHeight
  
    const startCloneEle = startElement.cloneNode(true)
    const endCloneEle = endElement.cloneNode(true)
  
    const startText = startElement.innerText.substring(0, startOffset)
    const endText = endElement.innerText.substring(endOffset)
  
    startCloneEle.innerText = startText
    endCloneEle.innerText = endText
    textLayer.appendChild(startCloneEle)
    textLayer.appendChild(endCloneEle)
  
    const startMatrix = getComputedStyle(startCloneEle).transform === 'none' ? null : getComputedStyle(startCloneEle).transform.split('(')[1].split(')')[0].split(', ')
    const endMatrix = getComputedStyle(endCloneEle).transform === 'none' ? null : getComputedStyle(endCloneEle).transform.split('(')[1].split(')')[0].split(', ')
    const startExtraWidth = startMatrix ? parseFloat(getComputedStyle(startCloneEle).width) * parseFloat(startMatrix[0]) : parseFloat(getComputedStyle(startCloneEle).width)
    const endExtraWidth = endMatrix ? parseFloat(getComputedStyle(endCloneEle).width) * parseFloat(endMatrix[0]) : parseFloat(getComputedStyle(endCloneEle).width)
  
    textLayer.removeChild(startCloneEle)
    textLayer.removeChild(endCloneEle)

    const position = []
    
    const startRectMatrix = getComputedStyle(startElement).transform === 'none' ? null : getComputedStyle(startElement).transform.split('(')[1].split(')')[0].split(', ')
    const endRectMatrix = getComputedStyle(endElement).transform === 'none' ? null : getComputedStyle(endElement).transform.split('(')[1].split(')')[0].split(', ')
    const startRectWidth = parseFloat(getComputedStyle(startElement).width) * (startRectMatrix ? parseFloat(startRectMatrix[0]) : 1)
    const endRectWidth = parseFloat(getComputedStyle(endElement).width) * (endRectMatrix ? parseFloat(endRectMatrix[0]) : 1)
    const startRectHeight = parseFloat(getComputedStyle(startElement).height) * (startRectMatrix ? parseFloat(startRectMatrix[3]) : 1)
    const endRectHeight = parseFloat(getComputedStyle(endElement).height) * (endRectMatrix ? parseFloat(endRectMatrix[3]) : 1)
  
    // left to right and up to down select
    let startX = startElement.offsetLeft + startExtraWidth
    let startY = startElement.offsetTop
    let endX = endElement.offsetLeft + endRectWidth - endExtraWidth
    let endY = endElement.offsetTop + endRectHeight
  
    if (startX > endX && startY >= endY) {
      // right to left and down to up select
      startX = endElement.offsetLeft + startExtraWidth
      startY = endElement.offsetTop
      endX = startElement.offsetLeft + startRectWidth - endExtraWidth
      endY = startElement.offsetTop + startRectHeight
    }
  
    const textElements = Array.prototype.slice.call(textLayer.childNodes)
  
    let inRange = false
  
    textElements.forEach((ele) => {
      if (!ele.innerText.trim()) return
      const rectMatrix = getComputedStyle(ele).transform === 'none' ? null : getComputedStyle(ele).transform.split('(')[1].split(')')[0].split(', ')
      const width = rectMatrix ? parseFloat(getComputedStyle(ele).width) * parseFloat(rectMatrix[0]) : parseFloat(getComputedStyle(ele).width)
      const height = rectMatrix ? parseFloat(getComputedStyle(ele).height) * parseFloat(rectMatrix[3]) : parseFloat(getComputedStyle(ele).height)
      const { offsetTop, offsetLeft } = ele
  
      const top = offsetTop
      const left = offsetLeft
      const right = offsetLeft + width
      const bottom = offsetTop + height
      const coords = {
        top,
        bottom,
        left,
        right,
      }
  
      if (startElement === ele && endElement === ele) {
        // in same element
        coords.left = startX
        coords.right = endX
        position.push(coords)
      } else if (startElement === ele) {
        // in start element
        coords.left = startX
        position.push(coords)
        inRange = true
      } else if (endElement === ele) {
        // in end element
        coords.right = endX
        position.push(coords)
        inRange = false
      } else if (inRange) {
        // middle element
        position.push(coords)
      }
    })
  
    for (let i = 1; i < position.length; i += 1) {
      if (
        position[i].top >= position[i - 1].top - 4 &&
        position[i].bottom <= position[i - 1].bottom + 4
      ) {
        position[i - 1].right = position[i].left
      }
    }
    const topArray = []
    const bottomArray = []
    const leftArray = []
    const rightArray = []
    const newPosition = position.reduce((acc, current) => {
      const actualPosition = this.parsePositionForBackend(current, pageHeight)
      topArray.push(actualPosition.top)
      bottomArray.push(actualPosition.bottom)
      leftArray.push(actualPosition.left)
      rightArray.push(actualPosition.right)
      if (!acc.length) {
        acc.push(actualPosition)
        return acc
      }
      const endAcc = acc[acc.length - 1]
      if (actualPosition.left === endAcc.right && actualPosition.top === endAcc.top && actualPosition.bottom === endAcc.bottom) {
        endAcc.right = actualPosition.right
      } else {
        acc.push(actualPosition)
      }
      return acc
    }, [])

    let quadPoints = ''
    for (let i = 0; i < newPosition.length; i++) {
      const curPosition = newPosition[i]
      if (i > 0) {
        quadPoints += ','
      }
      quadPoints += `${curPosition.left},${curPosition.top},${curPosition.right},${curPosition.top},${curPosition.left},${curPosition.bottom},${curPosition.right},${curPosition.bottom}`
    }

    const top = Math.max(...topArray)
    const bottom = Math.min(...bottomArray)
    const left = Math.min(...leftArray)
    const right = Math.max(...rightArray)

    const annotationData = {
      operate: "add-annot",
      type: this.tool,
      page: startPageNum - 1,
      transparency: 1,
      quadPoints,
      rect: `${left},${bottom},${right},${top}`,
      borderColor: this.color,
      content: content || undefined
    }
    if (this.tool === 'highlight') {
      annotationData.transparency = 0.5
    }
    return annotationData
  }
  parsePositionForBackend (position, height) {
    const positionArray = {
      left: Math.round(position.left * 1000 / this.scale) / 1000,
      top: Math.round((height - position.top) * 1000 / this.scale) / 1000,
      right: Math.round(position.right * 1000 / this.scale) / 1000,
      bottom: Math.round((height - position.bottom) * 1000 / this.scale) / 1000
    }

    return positionArray
  }
}
export default MarkupAnnotation
